@@ -3,8 +3,10 @@ package XML::Printer::ESCPOS::Tags;
 use strict;
 use warnings;
 use Text::Wrapper;
+use GD;
 
-our $VERSION = '0.04';
+
+our $VERSION = '0.05';
 
 =head2 new
 
@@ -37,13 +39,19 @@ sub tag_allowed {
         invert
         lf
         printAreaWidth
+        set
+        unset
         qr
+        repeat
         rot90
         tab
+        tabpositions
         text
         underline
         upsideDown
         utf8ImagedText
+        justify
+        hr
         /;
 }
 
@@ -90,6 +98,21 @@ sub simple_switch {
     return 1;
 }
 
+=head2 include_global_options
+
+Helper method to add global options to object options.
+
+=cut
+
+sub include_global_options {
+    my ($self, $options) = @_;
+
+    return {
+        %{ $self->{global_options} // {} },
+        %$options,
+    };
+}
+
 =head2 _0
 
 Prints plain text and strips out leading and trailing whitespaces.
@@ -100,7 +123,10 @@ sub _0 {
     my ( $self, $text ) = @_;
     $text =~ s/^\s+//gm;
     $text =~ s/\s+$//gm;
-    $self->{printer}->text($text) if $text =~ /\S/;
+    if ($text =~ /\S/) {
+        $self->{printer}->justify( $self->{justify_state} ) if $self->{justify_state};
+        $self->{printer}->text($text);
+    }
     return 1;
 }
 
@@ -118,16 +144,22 @@ sub _text {
     return $self->{caller}->_set_error_message("wrong text tag usage") if @$params != 3;
     return $self->{caller}->_set_error_message("wrong text tag usage") if ref $params->[0] ne 'HASH';
     return $self->{caller}->_set_error_message("wrong text tag usage") if $params->[1] != 0;
-    my $options = $params->[0];
+    my $options = $self->include_global_options($params->[0]);
     if ( exists $options->{wordwrap} ) {
         my $columns = delete $options->{wordwrap} || 49;
+        if ( $columns !~ /^\d+$/ or $columns < 1 ) {
+            return $self->{caller}->_set_error_message("wrong text tag usage: wordwrap attribute must be a positive integer");
+        }
+
         my $body_start = exists $options->{bodystart} ? delete $options->{bodystart} : '';
         my $wrapper = Text::Wrapper->new( columns => $columns, body_start => $body_start );
         for my $line ( split /\n/ => $wrapper->wrap( $params->[2] ) ) {
+            $self->{printer}->justify( $self->{justify_state} ) if $self->{justify_state};
             $self->{printer}->text($line);
         }
     }
     else {
+        $self->{printer}->justify( $self->{justify_state} ) if $self->{justify_state};
         $self->{printer}->text( $params->[2] );
     }
     return 1;
@@ -227,7 +259,7 @@ sub _qr {
     return $self->{caller}->_set_error_message("wrong QR code tag usage") if @$params != 3;
     return $self->{caller}->_set_error_message("wrong QR code tag usage") if ref $params->[0] ne 'HASH';
     return $self->{caller}->_set_error_message("wrong QR code tag usage") if $params->[1] != 0;
-    my $options = $params->[0];
+    my $options = $self->include_global_options($params->[0]);
     if (%$options) {
         $self->{printer}->qr( $params->[2], $options->{ecc} || 'L', $options->{version} || 5, $options->{moduleSize} || 3 );
     }
@@ -250,7 +282,7 @@ sub _barcode {
     return $self->{caller}->_set_error_message("wrong barcode tag usage") if @$params != 3;
     return $self->{caller}->_set_error_message("wrong barcode tag usage") if ref $params->[0] ne 'HASH';
     return $self->{caller}->_set_error_message("wrong barcode tag usage") if $params->[1] != 0;
-    my $options = $params->[0];
+    my $options = $self->include_global_options($params->[0]);
     if (%$options) {
         $self->{printer}->barcode( barcode => $params->[2], map { $_ => $options->{$_} } sort keys %$options );
     }
@@ -276,10 +308,15 @@ sub _utf8ImagedText {
     return $self->{caller}->_set_error_message("wrong utf8ImagedText tag usage") if @$params != 3;
     return $self->{caller}->_set_error_message("wrong utf8ImagedText tag usage") if ref $params->[0] ne 'HASH';
     return $self->{caller}->_set_error_message("wrong utf8ImagedText tag usage") if $params->[1] != 0;
-    my $options = $params->[0];
+    my $options = $self->include_global_options($params->[0]);
     if (%$options) {
         if ( exists $options->{wordwrap} ) {
             my $columns = delete $options->{wordwrap} || 49;
+            if ( $columns !~ /^\d+$/ or $columns < 1 ) {
+                return $self->{caller}
+                    ->_set_error_message("wrong utf8ImagedText tag usage: wordwrap attribute must be a positive integer");
+            }
+
             my $body_start = exists $options->{bodystart} ? delete $options->{bodystart} : '';
             my $wrapper = Text::Wrapper->new( columns => $columns, body_start => $body_start );
             for my $line ( split /\n/ => $wrapper->wrap( $params->[2] ) ) {
@@ -343,22 +380,41 @@ Print image from named file.
 sub _image {
     my ( $self, $params ) = @_;
 
+    my $filename;
+
     # single tag form <image filename="image.jpg" />
     if ( @$params == 1 ) {
         return $self->{caller}->_set_error_message("wrong image tag usage") if ref $params->[0] ne 'HASH';
         return $self->{caller}->_set_error_message("wrong image tag usage") if scalar keys %{ $params->[0] } != 1;
         return $self->{caller}->_set_error_message("wrong image tag usage") if not exists $params->[0]->{filename};
-        $self->{printer}->image( $params->[0]->{filename} );
-        return 1;
+        $filename = $params->[0]->{filename};
+    }
+    else {
+        # content tag form <image>image.jpg</image>
+        return $self->{caller}->_set_error_message("wrong image tag usage") if @$params != 3;
+        return $self->{caller}->_set_error_message("wrong image tag usage") if ref $params->[0] ne 'HASH';
+        return $self->{caller}->_set_error_message("wrong image tag usage") if %{ $params->[0] };
+        return $self->{caller}->_set_error_message("wrong image tag usage") if $params->[1] ne '0';
+        $filename = $params->[2];
     }
 
-    # content tag form <image>image.jpg</image>
-    return $self->{caller}->_set_error_message("wrong image tag usage") if @$params != 3;
-    return $self->{caller}->_set_error_message("wrong image tag usage") if ref $params->[0] ne 'HASH';
-    return $self->{caller}->_set_error_message("wrong image tag usage") if %{ $params->[0] };
-    return $self->{caller}->_set_error_message("wrong image tag usage") if $params->[1] ne '0';
+    return $self->{caller}->_set_error_message("wrong image tag usage: file does not exist") if !-f $filename;
 
-    $self->{printer}->image( $params->[2] );
+    my $image;
+    if ($filename =~ m/\.png$/) {
+        $image = GD::Image->newFromPng($filename) or return $self->{caller}->_set_error_message("Error loading image file $filename");
+    }
+    elsif ($filename =~ m/\.gif$/) {
+        $image = GD::Image->newFromGif($filename) or return $self->{caller}->_set_error_message("Error loading image file $filename");
+    }
+    elsif ($filename =~ m/\.jpe?g$/) {
+        $image = GD::Image->newFromJpeg($filename) or return $self->{caller}->_set_error_message("Error loading image file $filename");
+    }
+    else {
+        return $self->{caller}->_set_error_message("wrong image tag usage: file format not supported");
+    }
+
+    $self->{printer}->image( $image );
     return 1;
 }
 
@@ -377,6 +433,7 @@ sub _printAreaWidth {
         return $self->{caller}->_set_error_message("wrong printAreaWidth tag usage") if scalar keys %{ $params->[0] } != 1;
         return $self->{caller}->_set_error_message("wrong printAreaWidth tag usage") if not exists $params->[0]->{width};
         $self->{printer}->printAreaWidth( $params->[0]->{width} );
+        $self->{global_options}->{printAreaWidth} = $params->[0]->{width};
         return 1;
     }
 
@@ -387,16 +444,197 @@ sub _printAreaWidth {
     return $self->{caller}->_set_error_message("wrong printAreaWidth tag usage") if $params->[1] ne '0';
 
     $self->{printer}->printAreaWidth( $params->[2] );
+    $self->{global_options}->{printAreaWidth} = $params->[2];
     return 1;
 }
 
-=head2 _tabPositions
+=head2 _set
 
-Sets horizontal tab positions for tab stops.
+Sets global variables: paperWidth, wordwrap, fontFamily, fontSize, fontStyle, lineHeight, printAreaWidth
 
 =cut
 
-sub _tabPositions { }
+sub _set {
+    my ( $self, $params ) = @_;
+
+    return $self->{caller}->_set_error_message("wrong set tag usage") if @$params != 1 ;
+    return $self->{caller}->_set_error_message("wrong set tag usage") if ref $params->[0] ne 'HASH';
+    return $self->{caller}->_set_error_message("wrong set tag usage") if not scalar keys %{ $params->[0] };
+
+    $self->{global_options} ||= {};
+    for my $var (qw/paperWidth wordwrap fontFamily fontSize fontStyle lineHeight printAreaWidth/) {
+        $self->{global_options}->{$var} = $params->[0]->{$var} if $params->[0]->{$var};
+    }
+    
+    return 1;
+}
+
+=head2 _unset
+
+Resets global variables to standard values: paperWidth, wordwrap, fontFamily, fontSize, fontStyle, lineHeight, printAreaWidth
+Syntax: <unset fontStyle="" />
+
+=cut
+
+sub _unset {
+    my ( $self, $params ) = @_;
+
+    return $self->{caller}->_set_error_message("wrong unset tag usage") if @$params != 1 ;
+    return $self->{caller}->_set_error_message("wrong unset tag usage") if ref $params->[0] ne 'HASH';
+    return $self->{caller}->_set_error_message("wrong unset tag usage") if not scalar keys %{ $params->[0] };
+
+    for my $var (keys %{ $params->[0] }) {
+        delete $self->{global_options}->{$var};
+    }
+    
+    return 1;
+}
+
+=head2 _hr
+
+Adds a horizontal line. Use the I<thickness> attribute to set the line's thickness. Defaults to 2.
+
+=cut
+
+sub _hr {
+    my ( $self, $params ) = @_;
+    return $self->{caller}->_set_error_message("wrong hr tag usage") if @$params != 1;
+    return $self->{caller}->_set_error_message("wrong hr tag usage") if ref $params->[0] ne 'HASH';
+    my $thickness = 2;
+    if ( %{ $params->[0] } ) {
+        my @keys = keys %{ $params->[0] };
+        return $self->{caller}->_set_error_message("wrong hr tag usage") if !@keys;
+        $thickness = $params->[0]->{thickness};
+        return $self->{caller}->_set_error_message("wrong hr tag usage") if !$thickness;
+        return $self->{caller}->_set_error_message("wrong hr tag usage: thickness attribute must be a positive integer")
+            if $thickness !~ /^\d+$/ or $thickness < 1;
+    }
+
+    my $width = $params->[0]->{width} || $self->{paperWidth} || $self->{print_area_width} || 512;
+    my $img = GD::Image->new( $width, $thickness );
+    my $black = $img->colorAllocate( 0, 0, 0 );
+
+    for my $line ( 1 .. $thickness ) {
+        $img->line( 0, $line, $width, $line, $black );
+    }
+
+    $self->{printer}->image($img);
+
+    return 1;
+}
+
+=head2 _tabpositions
+
+Sets horizontal tab positions for tab stops. Syntax for XML is the following:
+
+    <tabpositions>
+      <tabposition>5</tabposition>
+      <tabposition>9</tabposition>
+      <tabposition>13</tabposition>
+    </tabpositions>
+
+=cut
+
+sub _tabpositions {
+    my ( $self, $tags ) = @_;
+
+    my @elements = @$tags;
+    my $hashref  = shift @elements;
+    if ( ref $hashref ne 'HASH' or %$hashref ) {
+        return $self->{caller}->_set_error_message('first element should be an empty hashref ({})');
+    }
+
+    my @tabpositions = ();
+    while (@elements) {
+        my $tag  = shift @elements;
+        my $data = shift @elements;
+        next if $tag eq '0' and $data =~ /^\s*$/;
+
+        if ( $tag ne 'tabposition' ) {
+            return $self->{caller}
+                ->_set_error_message("wrong tabpositions tag usage: must not contain anything else than tabposition tags");
+        }
+        if (   ref $data ne 'ARRAY'
+            or ref $data->[0] ne 'HASH'
+            or keys %{ $data->[0] }
+            or $data->[1] ne '0'
+            or $data->[2] !~ /^\d+$/
+            or $data->[2] < 1
+            or @$data != 3 )
+        {
+            return $self->{caller}->_set_error_message("wrong tabposition tag usage: value must be a positive integer");
+        }
+        push @tabpositions, $data->[2];
+    }
+
+    if ( !@tabpositions ) {
+        return $self->{caller}
+            ->_set_error_message("wrong tabpositions tag usage: must contain at least one tabposition tag as child");
+    }
+
+    $self->{printer}->tabPositions(@tabpositions);
+    return 1;
+}
+
+=head2 _repeat
+
+Repeats the content I<n> times. Syntax:
+
+    <repeat times="3">
+        [tags to repeat]
+    </repeat>
+
+=cut
+
+sub _repeat {
+    my ( $self, $params ) = @_;
+    return $self->{caller}->_set_error_message("wrong repeat tag usage") if ref $params ne 'ARRAY';
+    return $self->{caller}->_set_error_message("wrong repeat tag usage") if !@$params;
+    return $self->{caller}->_set_error_message("wrong repeat tag usage") if ref $params->[0] ne 'HASH';
+
+    my $times = 1;
+    my $options = $params->[0];
+    if ( exists $options->{times} ) {
+        $times = $options->{times};
+        return $self->{caller}->_set_error_message("wrong repeat tag usage: only positive integers are allowed") if $times =~ /\D/ or $times < 1;
+    }
+
+    for ( 1 .. $times ) {
+        $self->parse( [ {}, @$params[ 1.. $#$params ] ] ) or return;
+    }
+
+    return 1;
+}
+
+=head2 _justify
+
+Set justification to left, right or center. Syntax for XML is the following:
+<justify align="right">text</justify>
+
+=cut
+
+sub _justify {
+    my ( $self, $params ) = @_;
+
+    return $self->{caller}->_set_error_message("wrong justify tag usage") if @$params < 3;
+    return $self->{caller}->_set_error_message("wrong justify tag usage") if ref $params->[0] ne 'HASH';
+    return $self->{caller}->_set_error_message("wrong justify tag usage") if !$params->[0]->{align};
+    return $self->{caller}->_set_error_message("wrong justify tag usage") if keys %{ $params->[0] } != 1;
+    return $self->{caller}->_set_error_message("wrong justify tag usage") if not grep { $params->[0]->{align} eq $_ } qw/left center right/;
+
+    $self->{justify_state} ||= 'left';
+
+    my $justify_state_before = $self->{justify_state};
+    $self->{justify_state} = $params->[0]->{align};
+
+    $params->[0] = {};
+    $self->parse($params) or return;
+
+    $self->{justify_state} = $justify_state_before;
+    return 1;
+}
+
+=head1 NOT YET IMPLEMENTED TAGS
 
 =head2 _font
 
@@ -406,14 +644,6 @@ Choose font a, b or c.
 
 sub _font { }
 
-=head2 _justify
-
-Set justification to left, right or center.
-
-=cut
-
-sub _justify { }
-
 =head2 _fontHeight
 
 Set font height.
